from mirageml.classes import LoginManager
from mirageml.constants import REDIRECT_URI


def login():
    m = LoginManager(
        provider="google",
        provider_options={
            "redirect_to": REDIRECT_URI,
            "query_params": {
                "access_type": "offline",
                "prompt": "consent",
            },
        },
    )
    m.start_web_server()
    m.open_browser()



import json
import os

import typer


def show_config():
    config = load_config()
    from rich import print

    print(json.dumps(config, indent=4))


def load_config():
    config_path = os.path.expanduser("~/.mirageml.json")
    if os.path.exists(config_path):
        with open(config_path, "r") as f:
            return json.load(f)
    return {"local_mode": False, "model": "gpt-4"}


def save_config(config):
    config_path = os.path.expanduser("~/.mirageml.json")
    with open(config_path, "w") as f:
        json.dump(config, f, indent=4)


def set_config():
    config = load_config()

    valid = {
        "local_mode": (("True (not recommended without GPU)", "False"), json.loads),
        "model": (("gpt-3.5-turbo", "gpt-4"), str),
    }

    for key, value in config.items():
        if key not in valid:
            continue
        curvalue = value
        while True:
            question = f"Enter the value for '{key}' [{', '.join(valid[key][0])}] (current value: {curvalue}): "
            value = input(question)
            if value == "":
                break
            if value not in valid[key][0]:
                typer.secho(
                    f"Invalid value for '{key}'. Please use one of these options {valid[key][0]}",
                    fg=typer.colors.BRIGHT_RED,
                )
                continue
            config[key] = valid[key][1](value.lower())
            save_config(config)
            break

    typer.secho("MirageML Config updated!", fg=typer.colors.BRIGHT_GREEN)
    show_config()


def set_var_config(json_data):
    config = load_config()
    for key, value in json_data.items():
        config[key] = value
    save_config(config)



import os

import typer


def crawl_files(start_dir="."):
    file_data = []

    # Walk through the directory structure
    if os.path.isfile(start_dir):
        try:
            # Read the file content
            with open(start_dir, "r", encoding="utf-8") as file:
                file_content = file.read()
                file_data.append((file_content, start_dir))
        except Exception:
            # If unable to read a file, you can print an error or continue to the next file
            typer.secho(f"Unable to read file: {start_dir}", fg=typer.colors.BRIGHT_RED, bold=True)
    elif not os.path.isdir(start_dir):
        typer.secho(f"Unable to read dir: {start_dir}", fg=typer.colors.BRIGHT_RED, bold=True)
    else:
        for dirpath, dirnames, filenames in os.walk(os.path.abspath(start_dir)):
            for filename in filenames:
                # Skip hidden files
                if filename.startswith(".") or dirpath.split("/")[-1].startswith("."):
                    continue
                filepath = os.path.join(dirpath, filename)

                try:
                    # Read the file content
                    with open(filepath, "r", encoding="utf-8") as file:
                        file_content = file.read()
                        file_data.append((file_content, filepath))
                except Exception:
                    # If unable to read a file, you can print an error or continue to the next file
                    pass

    data = [x[1] + ": " + x[0] for x in file_data]
    metadata = [dict({"data": x[0]}, **{"source": x[1]}) for x in file_data]
    return data, metadata



"""
This file is the entry point for the commands module.
It initializes the module and imports the necessary modules.
"""
from .add_plugin import add_plugin
from .add_source import add_source
from .chat import chat
from .config import set_config, show_config
from .delete_source import delete_source
from .list_plugins import list_plugins
from .list_sources import list_sources
from .login import login
from .sync_plugin import sync_plugin

__all__ = [
    "login",
    "set_config",
    "show_config",
    "chat",
    "list_plugins",
    "list_sources",
    "add_plugin",
    "add_source",
    "delete_source",
    "sync_plugin",
]



import os
import sys
from io import StringIO

import requests

from ...constants import (
    LLM_GPT_ENDPOINT,
    get_headers,
)

PACKAGE_DIR = os.path.dirname(__file__)
os.environ["TRANSFORMERS_CACHE"] = os.path.join(PACKAGE_DIR, "models")


def _chunk_data(data, metadata):
    from langchain.text_splitter import RecursiveCharacterTextSplitter

    text_splitter = RecursiveCharacterTextSplitter(chunk_size=2048, chunk_overlap=80)
    docs = text_splitter.create_documents([data])
    chunks = [x.page_content for x in docs]
    meta = [{"data": curr_chunk, "source": metadata["source"]} for curr_chunk in chunks]
    vector_data = local_get_embedding(chunks)

    return chunks, meta, vector_data


def local_get_embedding(text_list, embedding_model_id="BAAI/bge-base-en-v1.5"):
    from sentence_transformers import SentenceTransformer

    os.environ["TOKENIZERS_PARALLELISM"] = "false"

    model_dir = os.path.join(PACKAGE_DIR, "models", embedding_model_id)
    if not os.path.exists(model_dir):
        os.makedirs(model_dir, exist_ok=True)
        print("Downloading model to:", model_dir)
        print("This will take a few minutes and only happen once!")

    # TODO: Suppress stdout/stderr
    original_stdout, original_stderr = sys.stdout, sys.stderr
    sys.stdout, sys.stderr = StringIO(), StringIO()

    model = SentenceTransformer(embedding_model_id, cache_folder=model_dir)
    embeddings = model.encode(text_list, normalize_embeddings=False)

    # Restore stdout/stderr
    sys.stdout, sys.stderr = original_stdout, original_stderr

    # Convert the embeddings to a list
    embeddings = embeddings.tolist()  # size = 768
    return embeddings


def local_llm_call(messages, llm_model_id="TheBloke/Llama-2-7b-Chat-GGUF", stream=False):
    from ctransformers import AutoModelForCausalLM

    os.environ["TOKENIZERS_PARALLELISM"] = "false"

    model_dir = os.path.join(PACKAGE_DIR, "models", llm_model_id)
    if not os.path.exists(model_dir):
        os.makedirs(model_dir, exist_ok=True)
        print("Downloading model to:", model_dir)
        print("This will take a few minutes and only happen once!")

    # Suppress stdout/stderr
    original_stdout, original_stderr = sys.stdout, sys.stderr
    sys.stdout, sys.stderr = StringIO(), StringIO()

    llm = AutoModelForCausalLM.from_pretrained(
        "TheBloke/Llama-2-7b-Chat-GGUF",
        model_file="llama-2-7b-chat.Q4_K_M.gguf",
        model_type="llama",
    )

    # Restore stdout/stderr
    sys.stdout, sys.stderr = original_stdout, original_stderr
    formatted_messages = "\n".join([x["content"] for x in messages])

    if stream:
        return llm(formatted_messages, stream=True)
    else:
        return llm(formatted_messages)


def get_embedding(text_list, model="BAAI/bge-small-en-v1.5"):
    raise NotImplementedError


def llm_call(messages, model="gpt-3.5-turbo", stream=False, local=False):
    if local:
        return local_llm_call(messages, stream=stream)
    json_data = {"model": model, "messages": messages, "stream": stream}
    return requests.post(LLM_GPT_ENDPOINT, json=json_data, headers=get_headers(), stream=stream)



import http.server
import platform
import subprocess
import sys
import threading
import urllib.parse

import jwt
import keyring
import requests
import segment.analytics as analytics
import typer

from mirageml.constants import (
    ANALYTICS_WRITE_KEY,
    NOTION_SYNC_ENDPOINT,
    PORT,
    SERVICE_ID,
    SUPABASE_KEY,
    SUPABASE_URL,
    supabase,
)

analytics.write_key = ANALYTICS_WRITE_KEY


class LoginManager:
    def __init__(self, handler="mirage_auth_handler", provider="google", provider_options={}):
        self._server = None
        self._thread = None
        self._handler = handler
        self._provider = provider
        self._provider_options = provider_options

    def start_web_server(self):
        """Kick off a thread for the local webserver."""
        th = threading.Thread(target=self._start_local_server)
        th.start()
        self._thread = th

    def select_handler(self):
        if self._handler == "mirage_auth_handler":
            return Handler
        elif self._handler == "google_auth_handler":
            return GoogleHandler
        elif self._handler == "notion_auth_handler":
            return NotionHandler

    def _start_local_server(self):
        self._server = http.server.HTTPServer(("localhost", PORT), self.select_handler())
        self._server.serve_forever()

    def open_browser(self):
        """Opens the browser to the login page."""
        # Waits for the server to start.
        while self._server is None:
            oauth_response = supabase.auth.sign_in_with_oauth(
                {
                    "provider": self._provider,
                    "options": self._provider_options,
                }
            )
            url = oauth_response.url
            system = platform.system()
            if system == "Darwin":
                cmd = ["open", url]
            elif system == "Linux":
                cmd = ["xdg-open", url]
            elif system == "Windows":
                cmd = ["cmd", "/c", "start", url.replace("&", "^&")]
            else:
                raise RuntimeError(f"Unsupported system: {system}")
            subprocess.run(cmd, check=True)


class Handler(http.server.SimpleHTTPRequestHandler):
    """Handle the response from accounts.google.com."""

    # Sketchy static variable to hold response.
    info = None

    def log_message(self, format, *args):
        return
        # return super().log_message(format, *args)

    def update_keyring(self, info):
        access_token = info["access_token"]
        decoded = jwt.decode(access_token, algorithms=["HS256"], options={"verify_signature": False})
        keyring.set_password(SERVICE_ID, "access_token", info["access_token"])
        keyring.set_password(SERVICE_ID, "refresh_token", info["refresh_token"])
        keyring.set_password(SERVICE_ID, "expires_at", info["expires_at"])
        keyring.set_password(SERVICE_ID, "user_id", decoded["sub"])
        keyring.set_password(SERVICE_ID, "email", decoded["email"])
        analytics.identify(
            decoded["sub"],
            {
                "email": decoded["email"],
            },
        )
        return access_token, decoded["sub"]

    def callback_handler(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        # Serve an HTML page with JavaScript to send the fragment to the server
        self.wfile.write(
            b"""
    <html>
        <body>
            <script>
                // Send the fragment to the server
                fetch('/capture_fragment?' + location.hash.substr(1))
                .then(() => {
                    document.body.innerHTML = 'All set, feel free to close this tab';
                    window.close();
                });
            </script>
        </body>
    </html>
    """
        )

    def capture_fragment_handler(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(b"All set, feel free to close this tab")
        # parse the fragment for the access token and refresh token
        fragment = self.path.split("?")[1]
        info = dict(kv.split("=") for kv in fragment.split("&"))
        self.update_keyring(info)
        sys.exit(0)

    def do_GET(self):
        if self.path.startswith("/callback"):
            self.callback_handler()
        elif self.path.startswith("/capture_fragment"):
            self.capture_fragment_handler()


class GoogleHandler(Handler):
    def capture_fragment_handler(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(b"All set, feel free to close this tab")
        # parse the fragment for the access token and refresh token
        fragment = self.path.split("?")[1]
        info = dict(kv.split("=") for kv in fragment.split("&"))
        access_token, user_id = self.update_keyring(info)

        provider_token = info["provider_token"]
        provider_refresh_token = urllib.parse.unquote(info["provider_refresh_token"])

        if provider_token and provider_refresh_token:
            params = {
                "user_id": user_id,
                "provider_token": provider_token,
                "provider_refresh_token": provider_refresh_token,
            }
            headers = {
                "apikey": SUPABASE_KEY,
                "Authorization": f"Bearer {access_token}",
                "Content-Type": "application/json",
                "Prefer": "resolution=merge-duplicates",
            }
            requests.post(
                f"{SUPABASE_URL}/rest/v1/user_google_tokens",
                json=params,
                headers=headers,
            )

        sys.exit(0)


class NotionHandler(Handler):
    def capture_fragment_handler(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(b"All set, feel free to close this tab")
        # parse the fragment for the access token and refresh token
        fragment = self.path.split("?")[1]
        info = dict(kv.split("=") for kv in fragment.split("&"))
        access_token, user_id = self.update_keyring(info)
        provider_token = info["provider_token"]
        if provider_token:
            keyring.set_password(SERVICE_ID, "notion_provider_token", info["provider_token"])
            params = {
                "user_id": user_id,
                "provider_token": provider_token,
            }
            headers = {
                "apikey": SUPABASE_KEY,
                "Authorization": f"Bearer {access_token}",
                "Content-Type": "application/json",
                "Prefer": "resolution=merge-duplicates",
            }
            requests.post(
                f"{SUPABASE_URL}/rest/v1/user_notion_tokens",
                json=params,
                headers=headers,
            )
            headers = {
                "Authorization": f"Bearer {access_token}",
            }
            sync_response = requests.post(NOTION_SYNC_ENDPOINT, json={}, headers=headers)
            sync_response_data = sync_response.json()
            if "error" in sync_response_data:
                typer.secho(
                    f"Error syncing notion: {sync_response_data['error']}",
                    fg=typer.colors.BRIGHT_RED,
                    bold=True,
                )
            else:
                typer.secho(
                    "Syncing notion triggered successfully. You will receive an email when the sync is complete.",
                    fg=typer.colors.BRIGHT_GREEN,
                    bold=True,
                )
        sys.exit(0)

    def do_GET(self):
        if self.path.startswith("/callback"):
            self.callback_handler()
        elif self.path.startswith("/capture_fragment"):
            self.capture_fragment_handler()



import typer
from rich.box import HORIZONTALS
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel

from .config import load_config
from .list_sources import get_sources
from .rag import rag_chat
from .utils.codeblocks import (
    add_indices_to_code_blocks,
    copy_code_to_clipboard,
    extract_code_from_markdown,
)
from .utils.custom_inputs import multiline_input
from .utils.llm import llm_call

console = Console()
config = load_config()


def chat(files: list[str] = [], urls: list[str] = [], sources: list[str] = []):
    # Beginning of the chat sequence
    transient_sources = []
    if files or urls or sources:
        index_local = False
        if "local" in sources:
            index_local = True
            sources.remove("local")

        local_sources, remote_sources = get_sources()
        all_sources = list(set(local_sources + remote_sources))
        for source in sources:
            if source not in all_sources:
                typer.secho(
                    f"Source: {source} does not exist. Please add it with `mirageml add source`",
                    fg=typer.colors.RED,
                    bold=True,
                )
                sources.remove(source)

        if index_local:
            from .add_source import add_local_source

            sources.append(add_local_source("local"))

        if files or urls:
            from .utils.local_source import crawl_files
            from .utils.web_source import extract_from_url

            with Live(
                Panel(
                    "Scraping Files and Urls",
                    title="[bold green]Assistant[/bold green]",
                    border_style="green",
                ),
                console=console,
                transient=True,
            ) as live:
                for source, extractor in zip(files + urls, [crawl_files] * len(files) + [extract_from_url] * len(urls)):
                    live.update(
                        Panel(
                            f"Extracting data from {source}",
                            title="[bold green]Assistant[/bold green]",
                            border_style="green",
                        )
                    )
                    data, metadata = extractor(source)
                    transient_sources.append((data, metadata))

    while True:
        chat_history = [{"role": "system", "content": "You are a helpful assistant."}]
        ai_response = ""
        if sources or transient_sources:
            chat_history, ai_response = rag_chat(sources, transient_sources)

        while True:
            # Loop for follow-up questions
            try:
                if ai_response:
                    chat_history.append({"role": "assistant", "content": ai_response})
                    indexed_ai_response = add_indices_to_code_blocks(ai_response)
                    console.print(
                        Panel(
                            Markdown(indexed_ai_response),
                            title="[bold blue]Assistant[/bold blue]",
                            box=HORIZONTALS,
                            border_style="blue",
                        )
                    )
                    code_blocks = extract_code_from_markdown(ai_response)
                    if code_blocks:
                        selected_indices = multiline_input(
                            "Enter the numbers of the code blocks you want to copy, separated by commas (e.g. '1,3,4') or ask a follow-up. Type reset to search again"
                        )
                        try:
                            selected_indices = list(map(int, selected_indices.split(",")))
                            copy_code_to_clipboard(code_blocks, selected_indices)
                            ai_response = ""
                            continue
                        except ValueError:
                            user_input = selected_indices
                    else:
                        user_input = multiline_input(
                            "Ask a follow-up. Type reset to search again. Ctrl+C to stop generating"
                        )
                else:
                    user_input = multiline_input("Chat with Mirage")
                if user_input.lower().strip() == "exit":
                    typer.secho("Ending chat. Goodbye!", fg=typer.colors.BRIGHT_GREEN, bold=True)
                    return
                elif (
                    user_input.lower().strip() == "reset"
                    or user_input.lower().strip() == "restart"
                    or user_input.lower().strip() == "start over"
                    or user_input.lower().strip() == "clear"
                ):
                    break

            except KeyboardInterrupt:
                typer.secho("Ending chat. Goodbye!", fg=typer.colors.BRIGHT_GREEN, bold=True)
                return

            chat_history.append({"role": "user", "content": user_input})

            try:
                with Live(
                    Panel(
                        "Assistant is thinking...",
                        title="[bold blue]Assistant[/bold blue]",
                        box=HORIZONTALS,
                        border_style="blue",
                    ),
                    console=console,
                    transient=True,
                    auto_refresh=True,
                    refresh_per_second=8,
                ) as live:
                    response = llm_call(
                        chat_history,
                        model=config["model"],
                        stream=True,
                        local=config["local_mode"],
                    )

                    ai_response = ""
                    if config["local_mode"]:
                        for chunk in response:
                            ai_response += chunk
                            live.update(
                                Panel(
                                    Markdown(ai_response),
                                    title="[bold blue]Assistant[/bold blue]",
                                    box=HORIZONTALS,
                                    border_style="blue",
                                )
                            )
                    else:
                        for chunk in response.iter_content(chunk_size=512):
                            if chunk:
                                decoded_chunk = chunk.decode("utf-8")
                                ai_response += decoded_chunk
                                live.update(
                                    Panel(
                                        Markdown(ai_response),
                                        title="[bold blue]Assistant[/bold blue]",
                                        box=HORIZONTALS,
                                        border_style="blue",
                                    )
                                )
            except KeyboardInterrupt:
                pass
